import java.io.Serializable;

/**
 * Klassen Spill implementerer interfacen (kontrakten) ISPill. Klassen
 * representerer en omgang av spillet Fire-P-Rad.
 * @author Trond Hellem B
 * modifisert av Khalid Azim Mughal, 2005
 */
public class Spill implements ISpill,Serializable{

  private ISpiller []spillere; //tabell med referanser til spillerene
  private ISpilleBrett spilleBrett; //referanse til spillebrettet
  private IBrikke sisteTrekk; // siste trekket som ble gjort
  private ISpiller vinner; // vinneren av spillet
  private boolean fulltSpilleBrett; //sant dersom spillebrettet er fullt av brikker
  private int hvemSinTur;//Hvilken spiller som skal gjre neste trekk

  /**
   * Konstruktr. Initialiserer spillet og oppretter spillebrettet
   * @param antallRekker Antall rekker spillebrettet skal ha
   * @param antallKolonner Antall kolonner spillebrettet skal ha
   * @param hvemStarter Spillernummeret til den spilleren som skal starte spillet
   */
  public Spill(int antallRekker, int antallKolonner, int hvemStarter){

    //Oppretter spillertabell og spillere
    spillere = new Spiller[2];
    spillere[0] = new Spiller(1);
    spillere[1] = new Spiller(2);

    //Sett hvilken spiller som skal ta det frste trekket
    if(hvemStarter==1 || hvemStarter==2)
      hvemSinTur=hvemStarter;
    else
      //Dersom ugyldig spillernummer, starter spiller 1 spillet
      hvemSinTur=1;

    //Oppretter et nytt spillebrett
    spilleBrett=new SpilleBrett(antallRekker, antallKolonner);

    fulltSpilleBrett=false;
  }

  /**
   * Gjr et trekk i den spesifikke kolonnen for den spesifiserte spilleren
   * @param kolonne Kolonnen brikken skal slippes ned i
   * @param spillerNummer Nummeret til spilleren som gjr trekket
   * @return int Rekkenummeret som trekket ble gjort i, ellers verdi < 0
   */
  public int gjrTrekk(int kolonne, int spillerNummer){
    int rekke=-1;

    //Sjekk at kolonnenummeret er gyldig og at rett spiller gjr trekket
    if(spilleBrett.gyldigKolonneNummer(kolonne) && spillerNummer==hvemSinTur){
      //Gjr trekket
      sisteTrekk=spilleBrett.leggTilBrikke(kolonne, spillere[spillerNummer-1]);

      //Hent rekkenummeret brikken havnet i
      rekke=sisteTrekk.hentRekke();

      //Sett hvilken spiller som skal gjre neste trekk
      if(hvemSinTur==1)
        hvemSinTur=2;
      else
        hvemSinTur=1;

      //Kontroller om trekket frte til en vinner
      sjekkOmFerdig();
    }

    //Returner rekkenummeret, dersom ugyldig trekk returneres verdien -1
    return rekke;
  }

  /**
   * Henter spilleren med det aktuelle nummeret
   * @param spillerNummer Nummeret til spilleren
   * @return ISpiller Spilleren med det spesifiserte nummeret
   */
  public ISpiller hentSpiller(int spillerNummer){
    if(spillerNummer==1 || spillerNummer==2)
      //Returner en referanse til spilleren
      return spillere[spillerNummer-1];
    else
      //Dersom feil spillernummer returneres en null referanse
      return null;
  }

  /**
   * @return ISpiller Spilleren som vant spillet, dersom noen har vunnet,
   * ellers null
   */
  public ISpiller hentVinner(){
    if(vinner!=null)
      return vinner;
    else
      return null;
  }

  /**
   * @return ISpiller Spilleren som gjorde det siste trekket
   */
  public ISpiller hentNvrendeSpiller(){
    //Sjekk at det er gjort minst et trekk
    if(sisteTrekk!=null)
      //Returner en referanse til spilleren
      return sisteTrekk.hentSpiller();
    else
      return null;
  }

  /**
   * Returnerer en referanse til spillebrettet
   * @return ISpilleBrett Spillebrettet som det spilles p
   */
  public ISpilleBrett hentSpilleBrett(){
    return spilleBrett;
  }

  /**
   * Sjekker om spillet er ferdig
   * @return boolean True dersom spillet er ferdig, ellers false
   */
  public boolean spillErFerdig(){
    //Spillet er ferdig dersom en spiller har vunnet eller dersom
    //spillebrettet er fullt
    return fulltSpilleBrett || vinner!=null;
  }

  /**
   * Ekstra hjelpemetode.
   * @return int Nummeret til den spilleren som skal gjre neste trekk
   */
  public int hvilkenSpillerSinTur(){
    return hvemSinTur;
  }

  /**
   * @return true dersom angitt brikken tilhrer spilleren som gjorde
   *              siste trekk.
   */
  private boolean sjekkGyldigBrikke(int i, int j, IBrikke[][] brikker){
    return brikker[i][j]!= null &&
           brikker[i][j].hentSpiller().hentSpillerNummer()==
           sisteTrekk.hentSpiller().hentSpillerNummer();
  }

  /**
   * Metode for  kontrollere om spillet er ferdig
   */
  private void sjekkOmFerdig(){
    /*
     Sjekker om siste trekk frte til en vinner. Kontrollerer brettet
     med utgangspunkt i det siste trekket som ble gjort (X).
     ----------------------
     |d        k        n.|
     |   i     o     o    |
     |      a  l  g       |
     |r  e  k  X  k  e    |
     |      a  o  g       |
     |   i     n     o    |
     |d        n        n.|
     |         e          |
     ----------------------
     Spillet vil ogs vre ferdig dersom brettet er fullt, selv om
     ingen har vunnet spillet.
     */

    //Sjekk kun dersom:
    //1) Spillebrettet ikke er fullt av brikker, og
    //2) Ingen allerede har vunnet, og
    //3) Det er gjort minst et trekk

    if(!fulltSpilleBrett && vinner==null && sisteTrekk!=null){

      //sjekk om siste trekk frte til fire p rad

      //Hent spillebrettet med alle brikkene
      IBrikke [][]brikker=spilleBrett.hentSpilleBrett();
      int antallRekker=brikker.length;
      int antallKolonner=brikker[0].length;

      //Tar utgangspunkt i rekken og kolonnen siste trekket ble
      //plassert
      int sisteRekke=sisteTrekk.hentRekke();
      int sisteKolonne=sisteTrekk.hentKolonne();

      //Har minst en p rad i hver retning
      int pRadIRekke=1;
      int pRadIKolonne=1;
      int pRadIDiagonal1=1;
      int pRadIDiagonal2=1;

      //finner antall p rad i rekken siste trekket havnet i

      for(int i=sisteKolonne+1;i<antallKolonner;i++) {
        if(!sjekkGyldigBrikke(sisteRekke, i, brikker))
          break;
        pRadIRekke++;
      }
      for(int i=sisteKolonne-1;i>=0;i--) {
        if(!sjekkGyldigBrikke(sisteRekke, i, brikker))
          break;
        pRadIRekke++;
      }

      //finner antall p rad i kolonnen trekket ble gjort

      if(pRadIRekke<4){
        //Trenger kun  sjekke under siste lagte brikke, fordi det
        //ikke vil vre noen brikker p oversiden av den.
        for(int i=sisteRekke-1;i>=0;i--) {
          if(!sjekkGyldigBrikke(i, sisteKolonne, brikker))
            break;
          pRadIKolonne++;
        }
      }

      //finner antall p rad i diagonal 1

      if(pRadIRekke<4 && pRadIKolonne<4){
        for(int i=sisteRekke+1,j=sisteKolonne+1;
            i<antallRekker && j<antallKolonner;
            i++,j++){
          if(!sjekkGyldigBrikke(i, j, brikker))
            break;
          pRadIDiagonal1++;
        }
        for(int i=sisteRekke-1,j=sisteKolonne-1;i>=0 && j>=0; i--,j--){
          if(!sjekkGyldigBrikke(i, j, brikker))
            break;
          pRadIDiagonal1++;
        }
      }

      //finner antall p rad i diagonal 2

      if(pRadIRekke<4 && pRadIKolonne<4 && pRadIDiagonal1<4){
        for(int i=sisteRekke+1,j=sisteKolonne-1;
            i<antallRekker && j>=0;
            i++,j--){
          if(!sjekkGyldigBrikke(i, j, brikker))
            break;
          pRadIDiagonal2++;
        }
        for(int i=sisteRekke-1,j=sisteKolonne+1;
            i>=0 && j<antallKolonner; i--,j++){
          if(!sjekkGyldigBrikke(i, j, brikker))
            break;
          pRadIDiagonal2++;
        }
      }

      //sjekk om vi har en vinner

      if(pRadIRekke>=4 || pRadIKolonne>=4 || pRadIDiagonal1>=4 ||
          pRadIDiagonal2>=4){

        //Det siste trekket frte til (minst) fire p rad
        //Setter spilleren som gjorde det siste trekket som vinner
        vinner=sisteTrekk.hentSpiller();
      }

      //fullt brett sjekk

      if(vinner==null){
        //sjekk om brettet er fullt
        //Vi antar at brettet er fullt til det motsatte er bevist
        boolean fullt=true;
        int i=antallRekker-1;//trenger bare  sjekke den verste rekken
        for(int j=0;j<antallKolonner && fullt;j++)
          if(brikker[i][j]==null)
            //Hvis vi finner et tomt felt p brettet kan
            //brettet ikke vre fullt
            fullt=false;
        fulltSpilleBrett=fullt;
      }
    }
  }
}
